#include <algorithm>

#include "flexpage.h"



/* Reset the flexpage using 'offset', being the file offset. */

void Flexpage::reset(offset_t offset)
{
    _counter = 0;

    /* By definition (see "Flexible-Sized Page Objects - Object-Orientation
       in Operation Systems"), flexpages are aligned to multiples of their
       size.

       The size of the flexpage depends on the amount of space around the
       accessed page. It cannot exceed the size of the memory region. */

    size = max_multiple(region->start, region->end, PAGE_SIZE);

    /* The base address of the flexpage is computed from the region start
       and flexpage size. It will be no lower than the region start.

       Sent flexpages may use higher bases due to receive window constraints,
       these being communicated by the "hot spot". */

    base_addr = round(region->start, size);

    /* Get the file offset for the base of the flexpage. This will be a
       multiple of the flexpage size for alignment purposes. */

    base_offset = trunc_offset(offset, size);

    /* The page being accessed is relative to the base.
       (This is transient information recording the initialising access
       details.) */

    page_offset = trunc(offset - base_offset, PAGE_SIZE);
    page_addr = base_addr + page_offset;
}

bool Flexpage::decrement()
{
    if (_counter)
    {
        _counter--;
        return _counter == 0;
    }
    else
        return 0;
}

void Flexpage::increment()
{
    _counter++;
}

void Flexpage::invalidate()
{
    _counter = 0;
}

bool Flexpage::valid()
{
    return _counter != 0;
}

/* Return whether the flexpage supports the given file 'position'. */

bool Flexpage::supports_position(offset_t position)
{
    return (base_offset <= position) && (position < (base_offset + size));
}

/* Upgrade the flags involved with this flexpage. This is used to track the
   maximal flags employed by the different pagers, with the result being used in
   unmap operations. */

void Flexpage::upgrade(flags_t flags)
{
    if (flags && (flags != _flags))
        _flags |= flags;
}

/* Return a "send" flexpage for an access to 'offset' by positioning it relative
   to 'hot_spot' for the receive flexpage window. */

SendFlexpage Flexpage::to_send(offset_t offset, address_t hot_spot,
                               flags_t flags, offset_t max_offset)
{
    /* The dataspace offset of the flexpage base is a multiple of the flexpage
       size. */

    offset_t receive_base_offset = trunc_offset(offset, size);

    /* The offset of the accessed page within the flexpage is constrained by the
       current flexpage size. */

    length_t page_offset = trunc(offset - receive_base_offset, PAGE_SIZE);

    /* The receive flexpage offset (hot spot) must be constrained to the
       flexpage, both the size and the start. */

    length_t receive_size;

    if (max_offset)
    {
        receive_size = trunc_multiple(max_offset - receive_base_offset, PAGE_SIZE);
        receive_size = std::min(size, receive_size);
    }
    else
        receive_size = size;

    if (!receive_size)
        return SendFlexpage(base_addr, page_order(0), flags);

    length_t receive_page_offset = hot_spot % receive_size;

    while ((receive_size > PAGE_SIZE) && (receive_page_offset != page_offset))
    {
        receive_size /= 2;
        receive_page_offset = hot_spot % receive_size;
    }

    /* The flexpage base address is adjusted using the difference in page
       offsets. Where the receive flexpage offset is constained further, the
       base address will be raised to become closer to the accessed page. */

    length_t adjustment = page_offset - receive_page_offset;

    return SendFlexpage(base_addr + adjustment, page_order(receive_size), flags);
}

/* Return a representation of the flexpage for unmapping. */

SendFlexpage Flexpage::to_unmap()
{
    return SendFlexpage(base_addr, page_order(size), _flags);
}

// vim: tabstop=4 expandtab shiftwidth=4
