#include "memory_utils.h"



/* Return page 'n' for the configured page size. */

length_t page(unsigned int n)
{
    return PAGE_SIZE * n;
}

/* Return the order of 'size', where 2 ** order yields the size. */

unsigned int page_order(length_t size)
{
    /* Count zeros from the left, stopping at the first set bit, using the width
       of the size value (in bits, starting with the width in bytes) to
       calculate the position of this bit and thus the order of the value. */

    return sizeof(unsigned long) * 8 - 1 - __builtin_clzl(size);
}

/* Return 'value' rounded up to the nearest 'increment'. */

length_t round(length_t value, length_t increment)
{
    return trunc(value + increment - 1, increment);
}

offset_t round_offset(offset_t value, length_t increment)
{
    return trunc_offset(value + increment - 1, increment);
}

/* Return 'value' rounded up to the nearest multiple of 'increment'. */

length_t round_multiple(length_t value, length_t increment)
{
    length_t last = increment;

    while (1)
    {
        if (value < increment)
            return round(value, last);

        last = increment;
        increment *= 2;
    }
}

/* Return 'value' rounded down (or truncated) to the nearest 'increment'. */

length_t trunc(length_t value, length_t increment)
{
    return (value / increment) * increment;
}

offset_t trunc_offset(offset_t value, length_t increment)
{
    return (value / increment) * increment;
}

/* Return 'value' rounded down (or truncated) to the nearest multiple of
   'increment'. */

length_t trunc_multiple(length_t value, length_t increment)
{
    length_t last = increment;

    while (1)
    {
        if (value < increment)
            return trunc(value, last);

        last = increment;
        increment *= 2;
    }
}

/* Find the maximum size aligned region within the region from 'start' to (but
   not including) 'end', with the given initial 'increment'. */

length_t max_multiple(length_t start, length_t end, length_t increment)
{
    /* The largest possible aligned region is derived from the region size. */

    length_t size = trunc_multiple(end - start, increment);

    /* Apply the alignment to the start. */

    length_t aligned = round(start, size);

    /* If the region is aligned, return the size. */

    if (aligned == start)
        return size;

    /* If the region is not aligned to the current size, recalculate the aligned
       size. */

    length_t aligned_size;

    do
    {
        aligned_size = trunc_multiple(end - aligned, increment);
        size /= 2;

        /* Determine whether a smaller alignment could yield a larger aligned
           size. */

        if (aligned_size >= size)
            return aligned_size;

        aligned = round(start, size);
    }
    while (aligned > start);

    return aligned_size;
}

// vim: tabstop=4 expandtab shiftwidth=4
